#     Author for this part of the script: Rich Kelley, rk5devmail@gmail.com, @RGKelley5
import os


def walk_checker(password, length=4, strict=True, loop=False):
    with open(f"{os.path.dirname(__file__)}/graph.py", "r") as fin:
        graph = eval(fin.read())

    result = False
    path_length = 1

    if strict and len(password) != length:
        return result

    for i in range(len(password)):
        current_letter = password[i]

        if i < len(password) - 1:
            next_letter = password[i + 1]

            if current_letter in graph:
                if loop:
                    if next_letter in graph[current_letter].values():
                        path_length += 1

                        if path_length == length:
                            result = True

                    else:
                        result = False
                        path_length = 1

                else:
                    if next_letter in graph[current_letter].values() and next_letter.lower() != current_letter.lower():
                        path_length += 1

                        if path_length == length:
                            result = True

                    else:
                        result = False
                        path_length = 1

    return result



{'!': {'diag_down_left': '=',
       'diag_down_right': 'w',
       'diag_up_left': '/',
       'diag_up_right': 'x',
       'down': 'q',
       'left': '`',
       'loop': '!',
       'right': '2',
       'shift_diag_down_left': '+',
       'shift_diag_down_right': 'W',
       'shift_diag_up_left': '?',
       'shift_diag_up_right': 'X',
       'shift_down': 'Q',
       'shift_left': '~',
       'shift_loop': '1',
       'shift_right': '@',
       'shift_up': 'Z',
       'up': 'z'},
 '"': {'diag_down_left': '/',
       'diag_down_right': '=',
       'diag_up_left': 'p',
       'diag_up_right': ']',
       'down': '-',
       'left': ';',
       'loop': '"',
       'right': 'a',
       'shift_diag_down_left': '?',
       'shift_diag_down_right': '+',
       'shift_diag_up_left': 'P',
       'shift_diag_up_right': '}',
       'shift_down': '_',
       'shift_left': ':',
       'shift_loop': "'",
       'shift_right': 'A',
       'shift_up': '{',
       'up': '['},
 '#': {'diag_down_left': 'w',
       'diag_down_right': 'r',
       'diag_up_left': 'x',
       'diag_up_right': 'v',
       'down': 'e',
       'left': '2',
       'loop': '#',
       'right': '4',
       'shift_diag_down_left': 'W',
       'shift_diag_down_right': 'R',
       'shift_diag_up_left': 'X',
       'shift_diag_up_right': 'V',
       'shift_down': 'E',
       'shift_left': '@',
       'shift_loop': '3',
       'shift_right': '$',
       'shift_up': 'C',
       'up': 'c'},
 '$': {'diag_down_left': 'e',
       'diag_down_right': 't',
       'diag_up_left': 'c',
       'diag_up_right': 'b',
       'down': 'r',
       'left': '3',
       'loop': '$',
       'right': '5',
       'shift_diag_down_left': 'E',
       'shift_diag_down_right': 'T',
       'shift_diag_up_left': 'C',
       'shift_diag_up_right': 'B',
       'shift_down': 'R',
       'shift_left': '#',
       'shift_loop': '4',
       'shift_right': '%',
       'shift_up': 'V',
       'up': 'v'},
 '%': {'diag_down_left': 'r',
       'diag_down_right': 'y',
       'diag_up_left': 'v',
       'diag_up_right': 'n',
       'down': 't',
       'left': '4',
       'loop': '%',
       'right': '6',
       'shift_diag_down_left': 'R',
       'shift_diag_down_right': 'Y',
       'shift_diag_up_left': 'V',
       'shift_diag_up_right': 'N',
       'shift_down': 'T',
       'shift_left': '$',
       'shift_loop': '5',
       'shift_right': '^',
       'shift_up': 'B',
       'up': 'b'},
 '&': {'diag_down_left': 'y',
       'diag_down_right': 'i',
       'diag_up_left': 'n',
       'diag_up_right': ',',
       'down': 'u',
       'left': '6',
       'loop': '&',
       'right': '8',
       'shift_diag_down_left': 'Y',
       'shift_diag_down_right': 'I',
       'shift_diag_up_left': 'N',
       'shift_diag_up_right': '<',
       'shift_down': 'U',
       'shift_left': '^',
       'shift_loop': '7',
       'shift_right': '*',
       'shift_up': 'M',
       'up': 'm'},
 "'": {'diag_down_left': '/',
       'diag_down_right': '=',
       'diag_up_left': 'p',
       'diag_up_right': ']',
       'down': '-',
       'left': ';',
       'loop': "'",
       'right': 'a',
       'shift_diag_down_left': '?',
       'shift_diag_down_right': '+',
       'shift_diag_up_left': 'P',
       'shift_diag_up_right': '}',
       'shift_down': '_',
       'shift_left': ':',
       'shift_loop': '"',
       'shift_right': 'A',
       'shift_up': '{',
       'up': '['},
 '(': {'diag_down_left': 'i',
       'diag_down_right': 'p',
       'diag_up_left': ',',
       'diag_up_right': '/',
       'down': 'o',
       'left': '8',
       'loop': '(',
       'right': '0',
       'shift_diag_down_left': 'I',
       'shift_diag_down_right': 'P',
       'shift_diag_up_left': '<',
       'shift_diag_up_right': '?',
       'shift_down': 'O',
       'shift_left': '*',
       'shift_loop': '9',
       'shift_right': ')',
       'shift_up': '>',
       'up': '.'},
 ')': {'diag_down_left': 'o',
       'diag_down_right': '[',
       'diag_up_left': '.',
       'diag_up_right': "'",
       'down': 'p',
       'left': '9',
       'loop': ')',
       'right': '-',
       'shift_diag_down_left': 'O',
       'shift_diag_down_right': '{',
       'shift_diag_up_left': '>',
       'shift_diag_up_right': '"',
       'shift_down': 'P',
       'shift_left': '(',
       'shift_loop': '0',
       'shift_right': '_',
       'shift_up': '?',
       'up': '/'},
 '*': {'diag_down_left': 'u',
       'diag_down_right': 'o',
       'diag_up_left': 'm',
       'diag_up_right': '.',
       'down': 'i',
       'left': '7',
       'loop': '*',
       'right': '9',
       'shift_diag_down_left': 'U',
       'shift_diag_down_right': 'O',
       'shift_diag_up_left': 'M',
       'shift_diag_up_right': '>',
       'shift_down': 'I',
       'shift_left': '&',
       'shift_loop': '8',
       'shift_right': '(',
       'shift_up': '<',
       'up': ','},
 '+': {'diag_down_left': '[',
       'diag_down_right': '\\',
       'diag_up_left': "'",
       'diag_up_right': '\\',
       'down': ']',
       'left': '-',
       'loop': '+',
       'right': '`',
       'shift_diag_down_left': '{',
       'shift_diag_down_right': '|',
       'shift_diag_up_left': '"',
       'shift_diag_up_right': '|',
       'shift_down': '}',
       'shift_left': '_',
       'shift_loop': '=',
       'shift_right': '~',
       'shift_up': '}',
       'up': ']'},
 ',': {'diag_down_left': '7',
       'diag_down_right': '9',
       'diag_up_left': 'j',
       'diag_up_right': 'l',
       'down': '8',
       'left': 'm',
       'loop': ',',
       'right': '.',
       'shift_diag_down_left': '&',
       'shift_diag_down_right': '(',
       'shift_diag_up_left': 'J',
       'shift_diag_up_right': 'L',
       'shift_down': '*',
       'shift_left': 'M',
       'shift_loop': '<',
       'shift_right': '>',
       'shift_up': 'K',
       'up': 'k'},
 '-': {'diag_down_left': 'p',
       'diag_down_right': ']',
       'diag_up_left': '/',
       'diag_up_right': ']',
       'down': '[',
       'left': '0',
       'loop': '-',
       'right': '=',
       'shift_diag_down_left': 'P',
       'shift_diag_down_right': '}',
       'shift_diag_up_left': '?',
       'shift_diag_up_right': '}',
       'shift_down': '{',
       'shift_left': ')',
       'shift_loop': '_',
       'shift_right': '+',
       'shift_up': '"',
       'up': "'"},
 '.': {'diag_down_left': '8',
       'diag_down_right': '0',
       'diag_up_left': 'k',
       'diag_up_right': ';',
       'down': '9',
       'left': ',',
       'loop': '.',
       'right': '/',
       'shift_diag_down_left': '*',
       'shift_diag_down_right': ')',
       'shift_diag_up_left': 'K',
       'shift_diag_up_right': ';',
       'shift_down': '(',
       'shift_left': '<',
       'shift_loop': '>',
       'shift_right': '?',
       'shift_up': 'L',
       'up': 'l'},
 '/': {'diag_down_left': '9',
       'diag_down_right': '-',
       'diag_up_left': 'l',
       'diag_up_right': "'",
       'down': '0',
       'left': '.',
       'loop': '/',
       'right': 'z',
       'shift_diag_down_left': '(',
       'shift_diag_down_right': '_',
       'shift_diag_up_left': 'L',
       'shift_diag_up_right': '"',
       'shift_down': ')',
       'shift_left': '>',
       'shift_loop': '?',
       'shift_right': 'Z',
       'shift_up': ':',
       'up': ';'},
 '0': {'diag_down_left': 'o',
       'diag_down_right': '[',
       'diag_up_left': '.',
       'diag_up_right': "'",
       'down': 'p',
       'left': '9',
       'loop': '0',
       'right': '-',
       'shift_diag_down_left': 'O',
       'shift_diag_down_right': '{',
       'shift_diag_up_left': '>',
       'shift_diag_up_right': '"',
       'shift_down': 'P',
       'shift_left': '(',
       'shift_loop': ')',
       'shift_right': '_',
       'shift_up': '?',
       'up': '/'},
 '1': {'diag_down_left': '=',
       'diag_down_right': 'w',
       'diag_up_left': '/',
       'diag_up_right': 'x',
       'down': 'q',
       'left': '`',
       'loop': '1',
       'right': '2',
       'shift_diag_down_left': '+',
       'shift_diag_down_right': 'W',
       'shift_diag_up_left': '?',
       'shift_diag_up_right': 'X',
       'shift_down': 'Q',
       'shift_left': '~',
       'shift_loop': '!',
       'shift_right': '@',
       'shift_up': 'Z',
       'up': 'z'},
 '2': {'diag_down_left': 'q',
       'diag_down_right': 'e',
       'diag_up_left': 'z',
       'diag_up_right': 'c',
       'down': 'w',
       'left': '1',
       'loop': '2',
       'right': '3',
       'shift_diag_down_left': 'Q',
       'shift_diag_down_right': 'E',
       'shift_diag_up_left': 'X',
       'shift_diag_up_right': 'C',
       'shift_down': 'W',
       'shift_left': '!',
       'shift_loop': '@',
       'shift_right': '#',
       'shift_up': 'X',
       'up': 'x'},
 '3': {'diag_down_left': 'w',
       'diag_down_right': 'r',
       'diag_up_left': 'x',
       'diag_up_right': 'v',
       'down': 'e',
       'left': '2',
       'loop': '3',
       'right': '4',
       'shift_diag_down_left': 'W',
       'shift_diag_down_right': 'R',
       'shift_diag_up_left': 'X',
       'shift_diag_up_right': 'V',
       'shift_down': 'E',
       'shift_left': '@',
       'shift_loop': '#',
       'shift_right': '$',
       'shift_up': 'C',
       'up': 'c'},
 '4': {'diag_down_left': 'e',
       'diag_down_right': 't',
       'diag_up_left': 'c',
       'diag_up_right': 'b',
       'down': 'r',
       'left': '3',
       'loop': '4',
       'right': '5',
       'shift_diag_down_left': 'E',
       'shift_diag_down_right': 'T',
       'shift_diag_up_left': 'C',
       'shift_diag_up_right': 'B',
       'shift_down': 'R',
       'shift_left': '#',
       'shift_loop': '$',
       'shift_right': '%',
       'shift_up': 'V',
       'up': 'v'},
 '5': {'diag_down_left': 'r',
       'diag_down_right': 'y',
       'diag_up_left': 'v',
       'diag_up_right': 'n',
       'down': 't',
       'left': '4',
       'loop': '5',
       'right': '6',
       'shift_diag_down_left': 'R',
       'shift_diag_down_right': 'Y',
       'shift_diag_up_left': 'V',
       'shift_diag_up_right': 'N',
       'shift_down': 'T',
       'shift_left': '$',
       'shift_loop': '%',
       'shift_right': '^',
       'shift_up': 'B',
       'up': 'b'},
 '6': {'diag_down_left': 't',
       'diag_down_right': 'u',
       'diag_up_left': 'b',
       'diag_up_right': 'm',
       'down': 'y',
       'left': '5',
       'loop': '6',
       'right': '7',
       'shift_diag_down_left': 'T',
       'shift_diag_down_right': 'U',
       'shift_diag_up_left': 'B',
       'shift_diag_up_right': 'M',
       'shift_down': 'Y',
       'shift_left': '%',
       'shift_loop': '^',
       'shift_right': '&',
       'shift_up': 'N',
       'up': 'n'},
 '7': {'diag_down_left': 'y',
       'diag_down_right': 'i',
       'diag_up_left': 'n',
       'diag_up_right': ',',
       'down': 'u',
       'left': '6',
       'loop': '7',
       'right': '8',
       'shift_diag_down_left': 'Y',
       'shift_diag_down_right': 'I',
       'shift_diag_up_left': 'N',
       'shift_diag_up_right': '<',
       'shift_down': 'U',
       'shift_left': '^',
       'shift_loop': '&',
       'shift_right': '*',
       'shift_up': 'M',
       'up': 'm'},
 '8': {'diag_down_left': 'u',
       'diag_down_right': 'o',
       'diag_up_left': 'm',
       'diag_up_right': '.',
       'down': 'i',
       'left': '7',
       'loop': '8',
       'right': '9',
       'shift_diag_down_left': 'U',
       'shift_diag_down_right': 'O',
       'shift_diag_up_left': 'M',
       'shift_diag_up_right': '>',
       'shift_down': 'I',
       'shift_left': '&',
       'shift_loop': '*',
       'shift_right': '(',
       'shift_up': '<',
       'up': ','},
 '9': {'diag_down_left': 'i',
       'diag_down_right': 'p',
       'diag_up_left': ',',
       'diag_up_right': '/',
       'down': 'o',
       'left': '8',
       'loop': '9',
       'right': '0',
       'shift_diag_down_left': 'I',
       'shift_diag_down_right': 'P',
       'shift_diag_up_left': '<',
       'shift_diag_up_right': '?',
       'shift_down': 'O',
       'shift_left': '*',
       'shift_loop': '(',
       'shift_right': ')',
       'shift_up': '>',
       'up': '.'},
 ':': {'diag_down_left': '.',
       'diag_down_right': '-',
       'diag_up_left': 'o',
       'diag_up_right': '[',
       'down': '/',
       'left': 'l',
       'loop': ':',
       'right': "'",
       'shift_diag_down_left': '>',
       'shift_diag_down_right': '_',
       'shift_diag_up_left': 'O',
       'shift_diag_up_right': '{',
       'shift_down': '?',
       'shift_left': 'L',
       'shift_loop': ';',
       'shift_right': '"',
       'shift_up': 'P',
       'up': 'p'},
 ';': {'diag_down_left': '.',
       'diag_down_right': '-',
       'diag_up_left': 'o',
       'diag_up_right': '[',
       'down': '/',
       'left': 'l',
       'loop': ';',
       'right': "'",
       'shift_diag_down_left': '>',
       'shift_diag_down_right': '_',
       'shift_diag_up_left': 'O',
       'shift_diag_up_right': '{',
       'shift_down': '?',
       'shift_left': 'L',
       'shift_loop': ':',
       'shift_right': '"',
       'shift_up': 'P',
       'up': 'p'},
 '<': {'diag_down_left': '7',
       'diag_down_right': '9',
       'diag_up_left': 'j',
       'diag_up_right': 'l',
       'down': '8',
       'left': 'm',
       'loop': '<',
       'right': '.',
       'shift_diag_down_left': '&',
       'shift_diag_down_right': '(',
       'shift_diag_up_left': 'J',
       'shift_diag_up_right': 'L',
       'shift_down': '*',
       'shift_left': 'M',
       'shift_loop': ',',
       'shift_right': '>',
       'shift_up': 'K',
       'up': 'k'},
 '=': {'diag_down_left': '[',
       'diag_down_right': '\\',
       'diag_up_left': "'",
       'diag_up_right': '\\',
       'down': ']',
       'left': '-',
       'loop': '=',
       'right': '`',
       'shift_diag_down_left': '{',
       'shift_diag_down_right': '|',
       'shift_diag_up_left': '"',
       'shift_diag_up_right': '|',
       'shift_down': '}',
       'shift_left': '_',
       'shift_loop': '+',
       'shift_right': '~',
       'shift_up': '}',
       'up': ']'},
 '>': {'diag_down_left': '8',
       'diag_down_right': '0',
       'diag_up_left': 'k',
       'diag_up_right': ';',
       'down': '9',
       'left': ',',
       'loop': '>',
       'right': '/',
       'shift_diag_down_left': '*',
       'shift_diag_down_right': ')',
       'shift_diag_up_left': 'K',
       'shift_diag_up_right': ';',
       'shift_down': '(',
       'shift_left': '<',
       'shift_loop': '.',
       'shift_right': '?',
       'shift_up': 'L',
       'up': 'l'},
 '?': {'diag_down_left': '9',
       'diag_down_right': '-',
       'diag_up_left': 'l',
       'diag_up_right': "'",
       'down': '0',
       'left': '.',
       'loop': '?',
       'right': 'z',
       'shift_diag_down_left': '(',
       'shift_diag_down_right': '_',
       'shift_diag_up_left': 'L',
       'shift_diag_up_right': '"',
       'shift_down': ')',
       'shift_left': '>',
       'shift_loop': '/',
       'shift_right': 'Z',
       'shift_up': ':',
       'up': ';'},
 '@': {'diag_down_left': 'q',
       'diag_down_right': 'e',
       'diag_up_left': 'z',
       'diag_up_right': 'c',
       'down': 'w',
       'left': '1',
       'loop': '@',
       'right': '3',
       'shift_diag_down_left': 'Q',
       'shift_diag_down_right': 'E',
       'shift_diag_up_left': 'X',
       'shift_diag_up_right': 'C',
       'shift_down': 'W',
       'shift_left': '!',
       'shift_loop': '2',
       'shift_right': '#',
       'shift_up': 'X',
       'up': 'x'},
 'A': {'diag_down_left': '=',
       'diag_down_right': 'x',
       'diag_up_left': '`',
       'diag_up_right': 'w',
       'down': 'z',
       'left': "'",
       'loop': 'A',
       'right': 's',
       'shift_diag_down_left': '+',
       'shift_diag_down_right': 'X',
       'shift_diag_up_left': '`',
       'shift_diag_up_right': 'W',
       'shift_down': 'Z',
       'shift_left': '"',
       'shift_loop': 'a',
       'shift_right': 'S',
       'shift_up': 'Q',
       'up': 'q'},
 'B': {'diag_down_left': '4',
       'diag_down_right': '6',
       'diag_up_left': 'f',
       'diag_up_right': 'h',
       'down': '5',
       'left': 'v',
       'loop': 'B',
       'right': 'n',
       'shift_diag_down_left': '$',
       'shift_diag_down_right': '^',
       'shift_diag_up_left': 'F',
       'shift_diag_up_right': 'H',
       'shift_down': '%',
       'shift_left': 'V',
       'shift_loop': 'b',
       'shift_right': 'N',
       'shift_up': 'G',
       'up': 'g'},
 'C': {'diag_down_left': '2',
       'diag_down_right': '4',
       'diag_up_left': 's',
       'diag_up_right': 'f',
       'down': '3',
       'left': 'x',
       'loop': 'C',
       'right': 'v',
       'shift_diag_down_left': '@',
       'shift_diag_down_right': '$',
       'shift_diag_up_left': 'S',
       'shift_diag_up_right': 'F',
       'shift_down': '#',
       'shift_left': 'X',
       'shift_loop': 'c',
       'shift_right': 'V',
       'shift_up': 'D',
       'up': 'd'},
 'D': {'diag_down_left': 'x',
       'diag_down_right': 'v',
       'diag_up_left': 'w',
       'diag_up_right': 'r',
       'down': 'c',
       'left': 's',
       'loop': 'D',
       'right': 'f',
       'shift_diag_down_left': 'X',
       'shift_diag_down_right': 'V',
       'shift_diag_up_left': 'W',
       'shift_diag_up_right': 'R',
       'shift_down': 'C',
       'shift_left': 'S',
       'shift_loop': 'd',
       'shift_right': 'F',
       'shift_up': 'E',
       'up': 'e'},
 'E': {'diag_down_left': 's',
       'diag_down_right': 'f',
       'diag_up_left': '2',
       'diag_up_right': '4',
       'down': 'd',
       'left': 'w',
       'loop': 'E',
       'right': 'r',
       'shift_diag_down_left': 'S',
       'shift_diag_down_right': 'F',
       'shift_diag_up_left': '@',
       'shift_diag_up_right': '$',
       'shift_down': 'D',
       'shift_left': 'W',
       'shift_loop': 'e',
       'shift_right': 'R',
       'shift_up': '#',
       'up': '3'},
 'F': {'diag_down_left': 'c',
       'diag_down_right': 'b',
       'diag_up_left': 'e',
       'diag_up_right': 't',
       'down': 'v',
       'left': 'd',
       'loop': 'F',
       'right': 'g',
       'shift_diag_down_left': 'C',
       'shift_diag_down_right': 'B',
       'shift_diag_up_left': 'E',
       'shift_diag_up_right': 'T',
       'shift_down': 'V',
       'shift_left': 'D',
       'shift_loop': 'f',
       'shift_right': 'G',
       'shift_up': 'R',
       'up': 'r'},
 'G': {'diag_down_left': 'v',
       'diag_down_right': 'n',
       'diag_up_left': 'r',
       'diag_up_right': 'y',
       'down': 'b',
       'left': 'f',
       'loop': 'G',
       'right': 'h',
       'shift_diag_down_left': 'V',
       'shift_diag_down_right': 'N',
       'shift_diag_up_left': 'R',
       'shift_diag_up_right': 'Y',
       'shift_down': 'B',
       'shift_left': 'F',
       'shift_loop': 'g',
       'shift_right': 'H',
       'shift_up': 'T',
       'up': 't'},
 'H': {'diag_down_left': 'b',
       'diag_down_right': 'm',
       'diag_up_left': 't',
       'diag_up_right': 'u',
       'down': 'n',
       'left': 'g',
       'loop': 'H',
       'right': 'j',
       'shift_diag_down_left': 'B',
       'shift_diag_down_right': 'M',
       'shift_diag_up_left': 'T',
       'shift_diag_up_right': 'U',
       'shift_down': 'N',
       'shift_left': 'G',
       'shift_loop': 'h',
       'shift_right': 'J',
       'shift_up': 'Y',
       'up': 'y'},
 'I': {'diag_down_left': 'j',
       'diag_down_right': 'l',
       'diag_up_left': '7',
       'diag_up_right': '9',
       'down': 'k',
       'left': 'u',
       'loop': 'I',
       'right': 'o',
       'shift_diag_down_left': 'J',
       'shift_diag_down_right': 'L',
       'shift_diag_up_left': '&',
       'shift_diag_up_right': '(',
       'shift_down': 'K',
       'shift_left': 'U',
       'shift_loop': 'i',
       'shift_right': 'O',
       'shift_up': '*',
       'up': '8'},
 'J': {'diag_down_left': 'n',
       'diag_down_right': ',',
       'diag_up_left': 'y',
       'diag_up_right': 'i',
       'down': 'm',
       'left': 'h',
       'loop': 'J',
       'right': 'k',
       'shift_diag_down_left': 'N',
       'shift_diag_down_right': '<',
       'shift_diag_up_left': 'Y',
       'shift_diag_up_right': 'I',
       'shift_down': 'M',
       'shift_left': 'H',
       'shift_loop': 'j',
       'shift_right': 'K',
       'shift_up': 'U',
       'up': 'u'},
 'K': {'diag_down_left': 'm',
       'diag_down_right': '.',
       'diag_up_left': 'u',
       'diag_up_right': 'o',
       'down': ',',
       'left': 'j',
       'loop': 'K',
       'right': 'l',
       'shift_diag_down_left': 'M',
       'shift_diag_down_right': '>',
       'shift_diag_up_left': 'U',
       'shift_diag_up_right': 'O',
       'shift_down': '<',
       'shift_left': 'J',
       'shift_loop': 'k',
       'shift_right': 'L',
       'shift_up': 'I',
       'up': 'i'},
 'L': {'diag_down_left': ',',
       'diag_down_right': '/',
       'diag_up_left': 'i',
       'diag_up_right': 'p',
       'down': '.',
       'left': 'k',
       'loop': 'L',
       'right': ';',
       'shift_diag_down_left': '<',
       'shift_diag_down_right': '?',
       'shift_diag_up_left': 'I',
       'shift_diag_up_right': 'P',
       'shift_down': '>',
       'shift_left': 'K',
       'shift_loop': 'l',
       'shift_right': ':',
       'shift_up': 'O',
       'up': 'o'},
 'M': {'diag_down_left': '6',
       'diag_down_right': '8',
       'diag_up_left': 'h',
       'diag_up_right': 'k',
       'down': '7',
       'left': 'n',
       'loop': 'M',
       'right': ',',
       'shift_diag_down_left': '^',
       'shift_diag_down_right': '*',
       'shift_diag_up_left': 'H',
       'shift_diag_up_right': 'K',
       'shift_down': '&',
       'shift_left': 'N',
       'shift_loop': 'm',
       'shift_right': '<',
       'shift_up': 'J',
       'up': 'j'},
 'N': {'diag_down_left': '5',
       'diag_down_right': '7',
       'diag_up_left': 'g',
       'diag_up_right': 'j',
       'down': '6',
       'left': 'b',
       'loop': 'N',
       'right': 'm',
       'shift_diag_down_left': '%',
       'shift_diag_down_right': '&',
       'shift_diag_up_left': 'G',
       'shift_diag_up_right': 'J',
       'shift_down': '^',
       'shift_left': 'B',
       'shift_loop': 'n',
       'shift_right': 'M',
       'shift_up': 'H',
       'up': 'h'},
 'O': {'diag_down_left': 'k',
       'diag_down_right': ';',
       'diag_up_left': '8',
       'diag_up_right': '0',
       'down': 'l',
       'left': 'i',
       'loop': 'O',
       'right': 'p',
       'shift_diag_down_left': 'K',
       'shift_diag_down_right': ':',
       'shift_diag_up_left': '*',
       'shift_diag_up_right': ')',
       'shift_down': 'L',
       'shift_left': 'I',
       'shift_loop': 'o',
       'shift_right': 'P',
       'shift_up': '(',
       'up': '9'},
 'P': {'diag_down_left': 'l',
       'diag_down_right': "'",
       'diag_up_left': '9',
       'diag_up_right': '-',
       'down': ';',
       'left': 'o',
       'loop': 'P',
       'right': '[',
       'shift_diag_down_left': 'L',
       'shift_diag_down_right': '"',
       'shift_diag_up_left': '(',
       'shift_diag_up_right': '_',
       'shift_down': ':',
       'shift_left': 'O',
       'shift_loop': 'p',
       'shift_right': '{',
       'shift_up': ')',
       'up': '0'},
 'Q': {'diag_down_left': '\\',
       'diag_down_right': 'S',
       'diag_up_left': '`',
       'diag_up_right': '2',
       'down': 'a',
       'left': '\\',
       'loop': 'Q',
       'right': 'w',
       'shift_diag_down_left': '|',
       'shift_diag_down_right': 'S',
       'shift_diag_up_left': '~',
       'shift_diag_up_right': '@',
       'shift_down': 'A',
       'shift_left': '|',
       'shift_loop': 'q',
       'shift_right': 'W',
       'shift_up': '!',
       'up': '1'},
 'R': {'diag_down_left': 'd',
       'diag_down_right': 'g',
       'diag_up_left': '3',
       'diag_up_right': '5',
       'down': 'f',
       'left': 'e',
       'loop': 'R',
       'right': 't',
       'shift_diag_down_left': 'D',
       'shift_diag_down_right': 'G',
       'shift_diag_up_left': '#',
       'shift_diag_up_right': '%',
       'shift_down': 'F',
       'shift_left': 'E',
       'shift_loop': 'r',
       'shift_right': 'T',
       'shift_up': '$',
       'up': '4'},
 'S': {'diag_down_left': 'z',
       'diag_down_right': 'c',
       'diag_up_left': 'q',
       'diag_up_right': 'e',
       'down': 'x',
       'left': 'a',
       'loop': 'S',
       'right': 'd',
       'shift_diag_down_left': 'Z',
       'shift_diag_down_right': 'C',
       'shift_diag_up_left': 'Q',
       'shift_diag_up_right': 'E',
       'shift_down': 'X',
       'shift_left': 'A',
       'shift_loop': 's',
       'shift_right': 'D',
       'shift_up': 'W',
       'up': 'w'},
 'T': {'diag_down_left': 'f',
       'diag_down_right': 'h',
       'diag_up_left': '4',
       'diag_up_right': '6',
       'down': 'g',
       'left': 'r',
       'loop': 'T',
       'right': 'y',
       'shift_diag_down_left': 'F',
       'shift_diag_down_right': 'H',
       'shift_diag_up_left': '$',
       'shift_diag_up_right': '^',
       'shift_down': 'G',
       'shift_left': 'R',
       'shift_loop': 't',
       'shift_right': 'Y',
       'shift_up': '%',
       'up': '5'},
 'U': {'diag_down_left': 'h',
       'diag_down_right': 'k',
       'diag_up_left': '6',
       'diag_up_right': '8',
       'down': 'j',
       'left': 'y',
       'loop': 'U',
       'right': 'i',
       'shift_diag_down_left': 'H',
       'shift_diag_down_right': 'K',
       'shift_diag_up_left': '^',
       'shift_diag_up_right': '*',
       'shift_down': 'J',
       'shift_left': 'Y',
       'shift_loop': 'u',
       'shift_right': 'I',
       'shift_up': '&',
       'up': '7'},
 'V': {'diag_down_left': '3',
       'diag_down_right': '5',
       'diag_up_left': 'd',
       'diag_up_right': 'g',
       'down': '4',
       'left': 'c',
       'loop': 'V',
       'right': 'b',
       'shift_diag_down_left': '#',
       'shift_diag_down_right': '%',
       'shift_diag_up_left': 'D',
       'shift_diag_up_right': 'G',
       'shift_down': '$',
       'shift_left': 'C',
       'shift_loop': 'v',
       'shift_right': 'B',
       'shift_up': 'F',
       'up': 'f'},
 'W': {'diag_down_left': 'a',
       'diag_down_right': 'd',
       'diag_up_left': '1',
       'diag_up_right': '3',
       'down': 's',
       'left': 'q',
       'loop': 'W',
       'right': 'e',
       'shift_diag_down_left': 'A',
       'shift_diag_down_right': 'D',
       'shift_diag_up_left': '!',
       'shift_diag_up_right': '#',
       'shift_down': 'S',
       'shift_left': 'Q',
       'shift_loop': 'w',
       'shift_right': 'E',
       'shift_up': '@',
       'up': '2'},
 'X': {'diag_down_left': '1',
       'diag_down_right': '3',
       'diag_up_left': 'a',
       'diag_up_right': 'd',
       'down': '2',
       'left': 'z',
       'loop': 'X',
       'right': 'c',
       'shift_diag_down_left': '!',
       'shift_diag_down_right': '#',
       'shift_diag_up_left': 'A',
       'shift_diag_up_right': 'D',
       'shift_down': '@',
       'shift_left': 'Z',
       'shift_loop': 'x',
       'shift_right': 'C',
       'shift_up': 'S',
       'up': 's'},
 'Y': {'diag_down_left': 'g',
       'diag_down_right': 'j',
       'diag_up_left': '5',
       'diag_up_right': '7',
       'down': 'h',
       'left': 't',
       'loop': 'Y',
       'right': 'u',
       'shift_diag_down_left': 'G',
       'shift_diag_down_right': 'J',
       'shift_diag_up_left': '%',
       'shift_diag_up_right': '&',
       'shift_down': 'H',
       'shift_left': 'T',
       'shift_loop': 'y',
       'shift_right': 'U',
       'shift_up': '^',
       'up': '6'},
 'Z': {'diag_down_left': '`',
       'diag_down_right': '2',
       'diag_up_left': '`',
       'diag_up_right': 's',
       'down': '1',
       'left': '/',
       'loop': 'Z',
       'right': 'x',
       'shift_diag_down_left': '~',
       'shift_diag_down_right': '@',
       'shift_diag_up_left': '~',
       'shift_diag_up_right': 'S',
       'shift_down': '!',
       'shift_left': '?',
       'shift_loop': 'z',
       'shift_right': 'X',
       'shift_up': 'A',
       'up': 'a'},
 '[': {'diag_down_left': ';',
       'diag_down_right': '\\',
       'diag_up_left': '0',
       'diag_up_right': '=',
       'down': "'",
       'left': 'p',
       'loop': '[',
       'right': ']',
       'shift_diag_down_left': ':',
       'shift_diag_down_right': '|',
       'shift_diag_up_left': ')',
       'shift_diag_up_right': '+',
       'shift_down': '"',
       'shift_left': 'P',
       'shift_loop': '{',
       'shift_right': '}',
       'shift_up': '_',
       'up': '-'},
 '\\': {'diag_down_left': '"',
        'diag_down_right': 'z',
        'diag_up_left': '_',
        'diag_up_right': '`',
        'down': '/',
        'left': ']',
        'loop': '\\',
        'right': 'q',
        'shift_diag_down_left': '"',
        'shift_diag_down_right': 'Z',
        'shift_diag_up_left': '+',
        'shift_diag_up_right': '~',
        'shift_down': '?',
        'shift_left': '}',
        'shift_loop': '|',
        'shift_right': 'Q',
        'shift_up': '+',
        'up': '='},
 ']': {'diag_down_left': "'",
       'diag_down_right': '\\',
       'diag_up_left': '-',
       'diag_up_right': 'z',
       'down': '"',
       'left': '[',
       'loop': ']',
       'right': '\\',
       'shift_diag_down_left': '"',
       'shift_diag_down_right': '\\',
       'shift_diag_up_left': '_',
       'shift_diag_up_right': 'Z',
       'shift_down': '"',
       'shift_left': '{',
       'shift_loop': '}',
       'shift_right': '|',
       'shift_up': '+',
       'up': '='},
 '^': {'diag_down_left': 't',
       'diag_down_right': 'u',
       'diag_up_left': 'b',
       'diag_up_right': 'm',
       'down': 'y',
       'left': '5',
       'loop': '^',
       'right': '7',
       'shift_diag_down_left': 'T',
       'shift_diag_down_right': 'U',
       'shift_diag_up_left': 'B',
       'shift_diag_up_right': 'M',
       'shift_down': 'Y',
       'shift_left': '%',
       'shift_loop': '6',
       'shift_right': '&',
       'shift_up': 'N',
       'up': 'n'},
 '_': {'diag_down_left': 'p',
       'diag_down_right': ']',
       'diag_up_left': '/',
       'diag_up_right': ']',
       'down': '[',
       'left': '0',
       'loop': '_',
       'right': '=',
       'shift_diag_down_left': 'P',
       'shift_diag_down_right': '}',
       'shift_diag_up_left': '?',
       'shift_diag_up_right': '}',
       'shift_down': '{',
       'shift_left': ')',
       'shift_loop': '-',
       'shift_right': '+',
       'shift_up': '"',
       'up': "'"},
 '`': {'diag_down_left': '/',
       'diag_down_right': 'z',
       'diag_up_left': '=',
       'diag_up_right': 'z',
       'down': 'z',
       'left': '=',
       'loop': '`',
       'right': '1',
       'shift_diag_down_left': 'Z',
       'shift_diag_down_right': 'Z',
       'shift_diag_up_left': '+',
       'shift_diag_up_right': 'Z',
       'shift_down': 'Z',
       'shift_left': '+',
       'shift_loop': '~',
       'shift_right': '!',
       'shift_up': 'Z',
       'up': 'z'},
 'a': {'diag_down_left': '=',
       'diag_down_right': 'x',
       'diag_up_left': '`',
       'diag_up_right': 'w',
       'down': 'z',
       'left': "'",
       'loop': 'a',
       'right': 's',
       'shift_diag_down_left': '+',
       'shift_diag_down_right': 'X',
       'shift_diag_up_left': '`',
       'shift_diag_up_right': 'W',
       'shift_down': 'Z',
       'shift_left': '"',
       'shift_loop': 'A',
       'shift_right': 'S',
       'shift_up': 'Q',
       'up': 'q'},
 'b': {'diag_down_left': '4',
       'diag_down_right': '6',
       'diag_up_left': 'f',
       'diag_up_right': 'h',
       'down': '5',
       'left': 'v',
       'loop': 'b',
       'right': 'n',
       'shift_diag_down_left': '$',
       'shift_diag_down_right': '^',
       'shift_diag_up_left': 'F',
       'shift_diag_up_right': 'H',
       'shift_down': '%',
       'shift_left': 'V',
       'shift_loop': 'B',
       'shift_right': 'N',
       'shift_up': 'G',
       'up': 'g'},
 'c': {'diag_down_left': '2',
       'diag_down_right': '4',
       'diag_up_left': 's',
       'diag_up_right': 'f',
       'down': '3',
       'left': 'x',
       'loop': 'c',
       'right': 'v',
       'shift_diag_down_left': '@',
       'shift_diag_down_right': '$',
       'shift_diag_up_left': 'S',
       'shift_diag_up_right': 'F',
       'shift_down': '#',
       'shift_left': 'X',
       'shift_loop': 'C',
       'shift_right': 'V',
       'shift_up': 'D',
       'up': 'd'},
 'd': {'diag_down_left': 'x',
       'diag_down_right': 'v',
       'diag_up_left': 'w',
       'diag_up_right': 'r',
       'down': 'c',
       'left': 's',
       'loop': 'd',
       'right': 'f',
       'shift_diag_down_left': 'X',
       'shift_diag_down_right': 'V',
       'shift_diag_up_left': 'W',
       'shift_diag_up_right': 'R',
       'shift_down': 'C',
       'shift_left': 'S',
       'shift_loop': 'D',
       'shift_right': 'F',
       'shift_up': 'E',
       'up': 'e'},
 'e': {'diag_down_left': 's',
       'diag_down_right': 'f',
       'diag_up_left': '2',
       'diag_up_right': '4',
       'down': 'd',
       'left': 'w',
       'loop': 'e',
       'right': 'r',
       'shift_diag_down_left': 'S',
       'shift_diag_down_right': 'F',
       'shift_diag_up_left': '@',
       'shift_diag_up_right': '$',
       'shift_down': 'D',
       'shift_left': 'W',
       'shift_loop': 'E',
       'shift_right': 'R',
       'shift_up': '#',
       'up': '3'},
 'f': {'diag_down_left': 'c',
       'diag_down_right': 'b',
       'diag_up_left': 'e',
       'diag_up_right': 't',
       'down': 'v',
       'left': 'd',
       'loop': 'f',
       'right': 'g',
       'shift_diag_down_left': 'C',
       'shift_diag_down_right': 'B',
       'shift_diag_up_left': 'E',
       'shift_diag_up_right': 'T',
       'shift_down': 'V',
       'shift_left': 'D',
       'shift_loop': 'F',
       'shift_right': 'G',
       'shift_up': 'R',
       'up': 'r'},
 'g': {'diag_down_left': 'v',
       'diag_down_right': 'n',
       'diag_up_left': 'r',
       'diag_up_right': 'y',
       'down': 'b',
       'left': 'f',
       'loop': 'g',
       'right': 'h',
       'shift_diag_down_left': 'V',
       'shift_diag_down_right': 'N',
       'shift_diag_up_left': 'R',
       'shift_diag_up_right': 'Y',
       'shift_down': 'B',
       'shift_left': 'F',
       'shift_loop': 'G',
       'shift_right': 'H',
       'shift_up': 'T',
       'up': 't'},
 'h': {'diag_down_left': 'b',
       'diag_down_right': 'm',
       'diag_up_left': 't',
       'diag_up_right': 'u',
       'down': 'n',
       'left': 'g',
       'loop': 'h',
       'right': 'j',
       'shift_diag_down_left': 'B',
       'shift_diag_down_right': 'M',
       'shift_diag_up_left': 'T',
       'shift_diag_up_right': 'U',
       'shift_down': 'N',
       'shift_left': 'G',
       'shift_loop': 'H',
       'shift_right': 'J',
       'shift_up': 'Y',
       'up': 'y'},
 'i': {'diag_down_left': 'j',
       'diag_down_right': 'l',
       'diag_up_left': '7',
       'diag_up_right': '9',
       'down': 'k',
       'left': 'u',
       'loop': 'i',
       'right': 'o',
       'shift_diag_down_left': 'J',
       'shift_diag_down_right': 'L',
       'shift_diag_up_left': '&',
       'shift_diag_up_right': '(',
       'shift_down': 'K',
       'shift_left': 'U',
       'shift_loop': 'I',
       'shift_right': 'O',
       'shift_up': '*',
       'up': '8'},
 'j': {'diag_down_left': 'n',
       'diag_down_right': ',',
       'diag_up_left': 'y',
       'diag_up_right': 'i',
       'down': 'm',
       'left': 'h',
       'loop': 'j',
       'right': 'k',
       'shift_diag_down_left': 'N',
       'shift_diag_down_right': '<',
       'shift_diag_up_left': 'Y',
       'shift_diag_up_right': 'I',
       'shift_down': 'M',
       'shift_left': 'H',
       'shift_loop': 'J',
       'shift_right': 'K',
       'shift_up': 'U',
       'up': 'u'},
 'k': {'diag_down_left': 'm',
       'diag_down_right': '.',
       'diag_up_left': 'u',
       'diag_up_right': 'o',
       'down': ',',
       'left': 'j',
       'loop': 'k',
       'right': 'l',
       'shift_diag_down_left': 'M',
       'shift_diag_down_right': '>',
       'shift_diag_up_left': 'U',
       'shift_diag_up_right': 'O',
       'shift_down': '<',
       'shift_left': 'J',
       'shift_loop': 'K',
       'shift_right': 'L',
       'shift_up': 'I',
       'up': 'i'},
 'l': {'diag_down_left': ',',
       'diag_down_right': '/',
       'diag_up_left': 'i',
       'diag_up_right': 'p',
       'down': '.',
       'left': 'k',
       'loop': 'l',
       'right': ';',
       'shift_diag_down_left': '<',
       'shift_diag_down_right': '?',
       'shift_diag_up_left': 'I',
       'shift_diag_up_right': 'P',
       'shift_down': '>',
       'shift_left': 'K',
       'shift_loop': 'L',
       'shift_right': ':',
       'shift_up': 'O',
       'up': 'o'},
 'm': {'diag_down_left': '6',
       'diag_down_right': '8',
       'diag_up_left': 'h',
       'diag_up_right': 'k',
       'down': '7',
       'left': 'n',
       'loop': 'm',
       'right': ',',
       'shift_diag_down_left': '^',
       'shift_diag_down_right': '*',
       'shift_diag_up_left': 'H',
       'shift_diag_up_right': 'K',
       'shift_down': '&',
       'shift_left': 'N',
       'shift_loop': 'M',
       'shift_right': '<',
       'shift_up': 'J',
       'up': 'j'},
 'n': {'diag_down_left': '5',
       'diag_down_right': '7',
       'diag_up_left': 'g',
       'diag_up_right': 'j',
       'down': '6',
       'left': 'b',
       'loop': 'n',
       'right': 'm',
       'shift_diag_down_left': '%',
       'shift_diag_down_right': '&',
       'shift_diag_up_left': 'G',
       'shift_diag_up_right': 'J',
       'shift_down': '^',
       'shift_left': 'B',
       'shift_loop': 'N',
       'shift_right': 'M',
       'shift_up': 'H',
       'up': 'h'},
 'o': {'diag_down_left': 'k',
       'diag_down_right': ';',
       'diag_up_left': '8',
       'diag_up_right': '0',
       'down': 'l',
       'left': 'i',
       'loop': 'o',
       'right': 'p',
       'shift_diag_down_left': 'K',
       'shift_diag_down_right': ':',
       'shift_diag_up_left': '*',
       'shift_diag_up_right': ')',
       'shift_down': 'L',
       'shift_left': 'I',
       'shift_loop': 'O',
       'shift_right': 'P',
       'shift_up': '(',
       'up': '9'},
 'p': {'diag_down_left': 'l',
       'diag_down_right': "'",
       'diag_up_left': '9',
       'diag_up_right': '-',
       'down': ';',
       'left': 'o',
       'loop': 'p',
       'right': '[',
       'shift_diag_down_left': 'L',
       'shift_diag_down_right': '"',
       'shift_diag_up_left': '(',
       'shift_diag_up_right': '_',
       'shift_down': ':',
       'shift_left': 'O',
       'shift_loop': 'P',
       'shift_right': '{',
       'shift_up': ')',
       'up': '0'},
 'q': {'diag_down_left': '\\',
       'diag_down_right': 'S',
       'diag_up_left': '`',
       'diag_up_right': '2',
       'down': 'a',
       'left': '\\',
       'loop': 'q',
       'right': 'w',
       'shift_diag_down_left': '|',
       'shift_diag_down_right': 'S',
       'shift_diag_up_left': '~',
       'shift_diag_up_right': '@',
       'shift_down': 'A',
       'shift_left': '|',
       'shift_loop': 'Q',
       'shift_right': 'W',
       'shift_up': '!',
       'up': '1'},
 'r': {'diag_down_left': 'd',
       'diag_down_right': 'g',
       'diag_up_left': '3',
       'diag_up_right': '5',
       'down': 'f',
       'left': 'e',
       'loop': 'r',
       'right': 't',
       'shift_diag_down_left': 'D',
       'shift_diag_down_right': 'G',
       'shift_diag_up_left': '#',
       'shift_diag_up_right': '%',
       'shift_down': 'F',
       'shift_left': 'E',
       'shift_loop': 'R',
       'shift_right': 'T',
       'shift_up': '$',
       'up': '4'},
 's': {'diag_down_left': 'z',
       'diag_down_right': 'c',
       'diag_up_left': 'q',
       'diag_up_right': 'e',
       'down': 'x',
       'left': 'a',
       'loop': 's',
       'right': 'd',
       'shift_diag_down_left': 'Z',
       'shift_diag_down_right': 'C',
       'shift_diag_up_left': 'Q',
       'shift_diag_up_right': 'E',
       'shift_down': 'X',
       'shift_left': 'A',
       'shift_loop': 'S',
       'shift_right': 'D',
       'shift_up': 'W',
       'up': 'w'},
 't': {'diag_down_left': 'f',
       'diag_down_right': 'h',
       'diag_up_left': '4',
       'diag_up_right': '6',
       'down': 'g',
       'left': 'r',
       'loop': 't',
       'right': 'y',
       'shift_diag_down_left': 'F',
       'shift_diag_down_right': 'H',
       'shift_diag_up_left': '$',
       'shift_diag_up_right': '^',
       'shift_down': 'G',
       'shift_left': 'R',
       'shift_loop': 'T',
       'shift_right': 'Y',
       'shift_up': '%',
       'up': '5'},
 'u': {'diag_down_left': 'h',
       'diag_down_right': 'k',
       'diag_up_left': '6',
       'diag_up_right': '8',
       'down': 'j',
       'left': 'y',
       'loop': 'u',
       'right': 'i',
       'shift_diag_down_left': 'H',
       'shift_diag_down_right': 'K',
       'shift_diag_up_left': '^',
       'shift_diag_up_right': '*',
       'shift_down': 'J',
       'shift_left': 'Y',
       'shift_loop': 'U',
       'shift_right': 'I',
       'shift_up': '&',
       'up': '7'},
 'v': {'diag_down_left': '3',
       'diag_down_right': '5',
       'diag_up_left': 'd',
       'diag_up_right': 'g',
       'down': '4',
       'left': 'c',
       'loop': 'v',
       'right': 'b',
       'shift_diag_down_left': '#',
       'shift_diag_down_right': '%',
       'shift_diag_up_left': 'D',
       'shift_diag_up_right': 'G',
       'shift_down': '$',
       'shift_left': 'C',
       'shift_loop': 'V',
       'shift_right': 'B',
       'shift_up': 'F',
       'up': 'f'},
 'w': {'diag_down_left': 'a',
       'diag_down_right': 'd',
       'diag_up_left': '1',
       'diag_up_right': '3',
       'down': 's',
       'left': 'q',
       'loop': 'w',
       'right': 'e',
       'shift_diag_down_left': 'A',
       'shift_diag_down_right': 'D',
       'shift_diag_up_left': '!',
       'shift_diag_up_right': '#',
       'shift_down': 'S',
       'shift_left': 'Q',
       'shift_loop': 'W',
       'shift_right': 'E',
       'shift_up': '@',
       'up': '2'},
 'x': {'diag_down_left': '1',
       'diag_down_right': '3',
       'diag_up_left': 'a',
       'diag_up_right': 'd',
       'down': '2',
       'left': 'z',
       'loop': 'x',
       'right': 'c',
       'shift_diag_down_left': '!',
       'shift_diag_down_right': '#',
       'shift_diag_up_left': 'A',
       'shift_diag_up_right': 'D',
       'shift_down': '@',
       'shift_left': 'Z',
       'shift_loop': 'X',
       'shift_right': 'C',
       'shift_up': 'S',
       'up': 's'},
 'y': {'diag_down_left': 'g',
       'diag_down_right': 'j',
       'diag_up_left': '5',
       'diag_up_right': '7',
       'down': 'h',
       'left': 't',
       'loop': 'y',
       'right': 'u',
       'shift_diag_down_left': 'G',
       'shift_diag_down_right': 'J',
       'shift_diag_up_left': '%',
       'shift_diag_up_right': '&',
       'shift_down': 'H',
       'shift_left': 'T',
       'shift_loop': 'Y',
       'shift_right': 'U',
       'shift_up': '^',
       'up': '6'},
 'z': {'diag_down_left': '`',
       'diag_down_right': '2',
       'diag_up_left': '`',
       'diag_up_right': 's',
       'down': '1',
       'left': '/',
       'loop': 'z',
       'right': 'x',
       'shift_diag_down_left': '~',
       'shift_diag_down_right': '@',
       'shift_diag_up_left': '~',
       'shift_diag_up_right': 'S',
       'shift_down': '!',
       'shift_left': '?',
       'shift_loop': 'Z',
       'shift_right': 'X',
       'shift_up': 'A',
       'up': 'a'},
 '{': {'diag_down_left': ';',
       'diag_down_right': '\\',
       'diag_up_left': '0',
       'diag_up_right': '=',
       'down': "'",
       'left': 'p',
       'loop': '{',
       'right': ']',
       'shift_diag_down_left': ':',
       'shift_diag_down_right': '|',
       'shift_diag_up_left': ')',
       'shift_diag_up_right': '+',
       'shift_down': '"',
       'shift_left': 'P',
       'shift_loop': '[',
       'shift_right': '}',
       'shift_up': '_',
       'up': '-'},
 '|': {'diag_down_left': '"',
       'diag_down_right': 'z',
       'diag_up_left': '_',
       'diag_up_right': '`',
       'down': '/',
       'left': ']',
       'loop': '|',
       'right': 'q',
       'shift_diag_down_left': '"',
       'shift_diag_down_right': 'Z',
       'shift_diag_up_left': '+',
       'shift_diag_up_right': '~',
       'shift_down': '?',
       'shift_left': '}',
       'shift_loop': '\\',
       'shift_right': 'Q',
       'shift_up': '+',
       'up': '='},
 '}': {'diag_down_left': "'",
       'diag_down_right': '\\',
       'diag_up_left': '-',
       'diag_up_right': 'z',
       'down': '"',
       'left': '[',
       'loop': '}',
       'right': '\\',
       'shift_diag_down_left': '"',
       'shift_diag_down_right': '\\',
       'shift_diag_up_left': '_',
       'shift_diag_up_right': 'Z',
       'shift_down': '"',
       'shift_left': '{',
       'shift_loop': ']',
       'shift_right': '|',
       'shift_up': '+',
       'up': '='},
 '~': {'diag_down_left': '/',
       'diag_down_right': 'z',
       'diag_up_left': '=',
       'diag_up_right': 'z',
       'down': 'z',
       'left': '=',
       'loop': '~',
       'right': '1',
       'shift_diag_down_left': 'Z',
       'shift_diag_down_right': 'Z',
       'shift_diag_up_left': '+',
       'shift_diag_up_right': 'Z',
       'shift_down': 'Z',
       'shift_left': '+',
       'shift_loop': '`',
       'shift_right': '!',
       'shift_up': 'Z',
       'up': 'z'}}


__author__ = """Aria Bagheri"""
__email__ = 'ariab9342@gmail.com'
__version__ = '1.1.0'

from enum import Enum

import pyhibp
from pyhibp import pwnedpasswords

from .walk_check import walk_checker

pyhibp.set_user_agent(ua="Python Password Security/1.1.0 (Library to check password security)")


class PasswordSecurityRequirement(Enum):
    USE_UPPERCASE = 1
    USE_LOWERCASE = 2
    USE_NUMBERS = 3
    USE_SPECIAL_CHARACTERS = 4
    MIN_16_CHARACTERS = 5
    NOT_PUBLICLY_KNOWN = 6
    NOT_KEYBOARD_WALK = 7
    NO_BANNED_WORDS = 8


PSR = PasswordSecurityRequirement


class PasswordSecurity:
    verifiers: set[PSR] = {
        PSR.USE_LOWERCASE,
        PSR.USE_UPPERCASE,
        PSR.USE_NUMBERS,
        PSR.USE_SPECIAL_CHARACTERS,
        PSR.MIN_16_CHARACTERS,
        PSR.NOT_PUBLICLY_KNOWN,
        PSR.NOT_KEYBOARD_WALK,
        PSR.NO_BANNED_WORDS,
    }
    banned_words: set[str] = set()

    def __init__(self, verifiers: set[PSR] = None, banned_words: set[str] = None):
        if verifiers:
            self.verifiers = verifiers
        if banned_words:
            self.banned_words = banned_words

    def verify_password(self, password: str, additional_banned_words: set[str] = None):
        banned_words = self.banned_words
        additional_banned_words and banned_words.update(additional_banned_words)

        is_safe = True
        is_safe &= PSR.NO_BANNED_WORDS not in self.verifiers or not self.contains_banned_words(password,
                                                                                               banned_words)
        is_safe &= PSR.MIN_16_CHARACTERS not in self.verifiers or self.is_password_long_enough(password)

        set_problems = self.get_character_set_missing_list(password)
        is_safe &= PSR.USE_NUMBERS not in self.verifiers or PSR.USE_NUMBERS not in set_problems
        is_safe &= PSR.USE_UPPERCASE not in self.verifiers or PSR.USE_UPPERCASE not in set_problems
        is_safe &= PSR.USE_LOWERCASE not in self.verifiers or PSR.USE_LOWERCASE not in set_problems
        is_safe &= PSR.USE_SPECIAL_CHARACTERS not in self.verifiers or PSR.USE_SPECIAL_CHARACTERS not in set_problems

        is_safe &= PSR.NOT_PUBLICLY_KNOWN not in self.verifiers or not self.is_password_public(password)
        is_safe &= PSR.NOT_KEYBOARD_WALK not in self.verifiers or not self.is_keyboard_walk(password)

        return is_safe

    def check_password_safety(self, password: str, additional_banned_words: set[str] = None) -> set[PSR]:
        banned_words = self.banned_words
        additional_banned_words and banned_words.update(additional_banned_words)

        problems = set()
        PSR.NO_BANNED_WORDS in self.verifiers and \
            self.contains_banned_words(password, banned_words) and \
            problems.add(PSR.NO_BANNED_WORDS)

        character_set_problems = self.get_character_set_missing_list(password)
        problems.update(character_set_problems.intersection(self.verifiers))

        PSR.NOT_PUBLICLY_KNOWN in self.verifiers and \
            self.is_password_public(password) and \
            problems.add(PSR.NOT_PUBLICLY_KNOWN)

        PSR.NOT_KEYBOARD_WALK in self.verifiers and \
            self.is_keyboard_walk(password) and \
            problems.add(PSR.NOT_KEYBOARD_WALK)
        return problems

    @staticmethod
    def contains_banned_words(password: str, banned_words: set[str]):
        return any(map(lambda x: x in password or password in x, banned_words))

    @staticmethod
    def is_password_long_enough(password: str):
        return len(password) >= 16

    @staticmethod
    def get_character_set_missing_list(password: str) -> set[PSR]:
        has_uppercase = has_lowercase = has_numbers = has_special_characters = False

        # https://owasp.org/www-community/password-special-characters
        special_characters = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"

        problems = set()

        for c in password:
            has_uppercase |= c.isupper()
            has_lowercase |= c.islower()
            has_numbers |= c.isdigit()
            has_special_characters |= c in special_characters

        not has_uppercase and problems.add(PSR.USE_UPPERCASE)
        not has_lowercase and problems.add(PSR.USE_LOWERCASE)
        not has_numbers and problems.add(PSR.USE_NUMBERS)
        not has_special_characters and problems.add(PSR.USE_SPECIAL_CHARACTERS)

        return problems

    @staticmethod
    def is_password_public(password: str):
        return pwnedpasswords.is_password_breached(password) != 0

    @staticmethod
    def is_keyboard_walk(password: str):
        return walk_checker(password, strict=False)
